# reused and modified from simon0302010/MetaView

import json
import os
import subprocess


def check_installed():
    try:
        result = subprocess.run(
            ["exiftool", "-ver"],
            capture_output=True,
            text=True,
            check=True
        )
        return result.returncode == 0
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False

def get_metadata(file_path):
    result = subprocess.run(
        ["exiftool", "-j", "-b", file_path], capture_output=True, text=True
    )
    metadata = json.loads(result.stdout)
    return metadata[0] if metadata else {}


def write_metadata(file_path, new_data):
    if len(new_data) and os.path.exists(file_path):
        command = ["exiftool"]
        for key in new_data:
            command.append(f"-{key}={new_data[key]}")
        command.append(file_path)
        result = subprocess.run(command, capture_output=True, text=True)
        return result.stdout.strip()
    else:
        return "Nothing to write."

# TODO: deletion of permanent data
def delete_metadata(file_path, all=True, properties=[]):
    if os.path.exists(file_path):
        if all:
            command = ["exiftool", "-f", "-All=", file_path]
            result = subprocess.run(command, capture_output=True, text=True)
            os.remove(file_path + "_original")
            return result.stdout.strip()
        elif properties:
            command = ["exiftool", "-f"]
            for property in properties:
                command.append(f"-{property}=")
            command.append(file_path)
            result = subprocess.run(command, capture_output=True, text=True)
            try:
                os.remove(file_path + "_original")
            except FileNotFoundError:
                pass
            if result.stderr.strip():
                return False
            return result.stdout.strip()



import fnmatch

METADATA_OPTIONS = METADATA_OPTIONS = {
    "Everything (All Metadata)": None,
    "File System Information": [
        "FileModifyDate",
        "FileAccessDate",
        "RelatedSoundFile",
        "OriginalFileName",
        "ThumbnailFileName",
        "RawFileName",
    ],
    "Core Image Properties": [
        "ExifImageWidth",
        "ExifImageHeight",
        "Orientation",
        "SamplesPerPixel",
        "PhotometricInterpretation",
        "ExifByteOrder",
        "JFIFVersion",
        "ComponentsConfiguration",
        "ColorComponents",
    ],
    "Camera and Exposure Settings (EXIF)": [
        "Make",
        "Model",
        "Software",
        "ModifyDate",
        "DateTimeOriginal",
        "CreateDate",
        "ExposureTime",
        "FNumber",
        "ISO",
        "ShutterSpeedValue",
        "ApertureValue",
        "ExposureCompensation",
        "ExposureProgram",
        "MeteringMode",
        "Flash",
        "LightSource",
        "SubSecTime",
        "SubSecTimeOriginal",
        "SubSecTimeDigitized",
        "SubSecCreateDate",
        "SubSecDateTimeOriginal",
        "SubSecModifyDate",
        "FlashpixVersion",
        "ColorSpace",
        "CustomRendered",
        "ExposureMode",
        "WhiteBalance",
        "SceneCaptureType",
        "DigitalZoomRatio",
        "GainControl",
        "Contrast",
        "Saturation",
        "Sharpness",
        "SubjectDistanceRange",
        "SensingMethod",
        "FileSource",
        "SceneType",
        "MaxApertureValue",
        "BrightnessValue",
        "HDRImageType",
        "SensitivityType",
        "ISOSpeed",
    ],
    "Lens and Focus Details": [
        "FocalLength",
        "FocalLengthIn35mmFormat",
        "FocalPlaneXResolution",
        "FocalPlaneYResolution",
        "FocalPlaneResolutionUnit",
        "SubjectArea",
        "SubjectDistance",
        "LensInfo",
        "LensMake",
        "LensModel",
        "LensID",
        "Lens",
        "Lens35efl",
        "FocusMode",
        "AFPoint",
        "FocusRange",
    ],
    "Calculated Photographic Properties": [
        "ScaleFactor35efl",
        "ShutterSpeed",
        "Aperture",
        "CircleOfConfusion",
        "DOF",
        "FOV",
        "FocalLength35efl",
        "HyperfocalDistance",
        "LightValue",
    ],
    "GPS Geolocation Data": ["GPS*"],
    "Content and Authorship (IPTC & XMP)": [
        "ImageDescription",
        "Artist",
        "Copyright",
        "UserComment",
        "CreatorTool",
        "OwnerName",
        "Headline",
        "Instructions",
        "AuthorsPosition",
        "Credit",
        "Source",
        "City",
        "State",
        "Country",
        "TransmissionReference",
        "DateCreated",
        "TimeCreated",
        "CaptionWriter",
        "Title",
        "Rights",
        "Comment",
        "XPAuthor",
        "XPComment",
        "XPKeywords",
        "XPSubject",
        "XPTitle",
        "Rating",
        "CopyrightNotice",
        "By-line",
        "ObjectName",
        "Keywords",
        "Caption-Abstract",
        "Description",
        "Creator",
        "Subject",
        "UsageTerms",
        "CreatorWorkEmail",
        "HierarchicalSubject",
        "Label",
        "WebStatement",
    ],
    "Adobe XMP Metadata": [
        "XMPToolkit",
        "Format",
        "ColorMode",
        "ICCProfileName",
        "TextLayerName",
        "TextLayerText",
        "DocumentAncestors",
        "InstanceID",
        "DerivedFromInstanceID",
        "DerivedFromDocumentID",
        "DocumentID",
        "OriginalDocumentID",
        "HistoryAction",
        "HistoryInstanceID",
        "HistoryWhen",
        "HistorySoftwareAgent",
        "HistoryChanged",
        "HistoryParameters",
        "Marked",
        "VersionID",
        "Logging",
        "Channel",
        "SubFolder",
        "LastKeywordXMP",
        "LastKeywordIPTC",
        "NativeDigest",
        "About",
        "Creatortool",
        "OffsetSchema",
    ],
    "Embedded Thumbnail & Preview": [
        "ThumbnailImage",
        "ThumbnailOffset",
        "ThumbnailLength",
        "PreviewImage",
        "PreviewImageValid",
        "PreviewImageStart",
        "PreviewImageLength",
    ],
    "ICC Color Profile": [
        "ProfileCMMType",
        "ProfileVersion",
        "ProfileClass",
        "ColorSpaceData",
        "ProfileConnectionSpace",
        "ProfileDateTime",
        "ProfileFileSignature",
        "PrimaryPlatform",
        "CMMFlags",
        "DeviceManufacturer",
        "DeviceModel",
        "DeviceAttributes",
        "RenderingIntent",
        "ConnectionSpaceIlluminant",
        "ProfileCreator",
        "ProfileID",
        "ProfileCopyright",
        "ProfileDescription",
        "MediaWhitePoint",
        "MediaBlackPoint",
        "RedMatrixColumn",
        "GreenMatrixColumn",
        "BlueMatrixColumn",
        "DeviceMfgDesc",
        "DeviceModelDesc",
        "ViewingCondDesc",
        "ViewingCondIlluminant",
        "ViewingCondSurround",
        "ViewingCondIlluminantType",
        "Luminance",
        "MeasurementObserver",
        "MeasurementBacking",
        "MeasurementGeometry",
        "MeasurementFlare",
        "MeasurementIlluminant",
        "Technology",
        "RedTRC",
        "GreenTRC",
        "BlueTRC",
        "ProfileDescriptionML*",
        "ChromaticAdaptation",
    ],
    "Technical Image Composition": [
        "Compression",
        "EncodingProcess",
        "BitsPerSample",
        "YCbCrSubSampling",
        "YCbCrPositioning",
        "XResolution",
        "YResolution",
        "ResolutionUnit",
        "CompressedBitsPerPixel",
        "PlanarConfiguration",
        "StripOffsets",
        "RowsPerStrip",
        "StripByteCounts",
        "JPEGInterchangeFormat*",
        "Padding",
    ],
    "Multi-Picture Format (MPF)": [
        "MPFVersion",
        "NumberOfImages",
        "MPImageFlags",
        "MPImageFormat",
        "MPImageType",
        "MPImageLength",
        "MPImageStart",
        "DependentImage1EntryNumber",
        "DependentImage2EntryNumber",
    ],
    "Canon MakerNotes": [
        "MacroMode",
        "SelfTimer",
        "Quality",
        "CanonFlashMode",
        "ContinuousDrive",
        "RecordMode",
        "CanonImageSize",
        "EasyMode",
        "DigitalZoom",
        "Contrast",
        "Saturation",
        "Sharpness",
        "CameraISO",
        "FocusRange",
        "CanonExposureMode",
        "LensType",
        "MaxFocalLength",
        "MinFocalLength",
        "FocalUnits",
        "MaxAperture",
        "MinAperture",
        "FlashActivity",
        "FlashBits",
        "FocusContinuous",
        "AESetting",
        "ImageStabilization",
        "ZoomSourceWidth",
        "ZoomTargetWidth",
        "SpotMeteringMode",
        "ManualFlashOutput",
        "AutoISO",
        "BaseISO",
        "MeasuredEV",
        "TargetAperture",
        "TargetExposureTime",
        "SlowShutter",
        "SequenceNumber",
        "OpticalZoomCode",
        "FlashGuideNumber",
        "FlashExposureComp",
        "AutoExposureBracketing",
        "AEBBracketValue",
        "ControlMode",
        "FocusDistanceUpper",
        "FocusDistanceLower",
        "BulbDuration",
        "CameraType",
        "AutoRotate",
        "NDFilter",
        "SelfTimer2",
        "FlashOutput",
        "CanonImageType",
        "CanonFirmwareVersion",
        "FileNumber",
        "CanonModelID",
        "DateStampMode",
        "MyColorMode",
        "FirmwareRevision",
        "Categories",
        "IntelligentContrast",
        "ImageUniqueID",
        "FacesDetected",
        "TimeZone",
        "TimeZoneCity",
        "DaylightSavings",
        "AFAreaMode",
        "NumAFPoints",
        "ValidAFPoints",
        "CanonImageWidth",
        "CanonImageHeight",
        "AFImageWidth",
        "AFImageHeight",
        "AFAreaWidths",
        "AFAreaHeights",
        "AFAreaXPositions",
        "AFAreaYPositions",
        "AFPointsInFocus",
        "PrimaryAFPoint",
        "AspectRatio",
        "CroppedImageWidth",
        "CroppedImageHeight",
        "CroppedImageLeft",
        "CroppedImageTop",
        "VRDOffset",
    ],
    "Nikon MakerNotes": [
        "MakerNoteVersion",
        "ColorMode",
        "Quality",
        "WhiteBalance",
        "FlashSetting",
        "FlashType",
        "ISOSelection",
        "ImageAdjustment",
        "ToneComp",
        "AuxiliaryLens",
        "AFAreaMode",
        "AFPoint",
        "AFPointsInFocus",
        "SceneMode",
        "DistortionControl",
        "SaturationAdj",
        "NoiseReduction",
        "RetouchHistory",
        "ImageStabilization",
        "PictureControl*",
        "ActiveD-Lighting",
    ],
    "Olympus MakerNotes": [
        "SpecialMode",
        "CameraID",
        "EquipmentVersion",
        "CameraType2",
        "SerialNumber",
        "InternalSerialNumber",
        "FocalPlaneDiagonal",
        "BodyFirmwareVersion",
        "AELock",
        "MacroMode",
        "FocusProcess",
        "AFSearch",
        "AFAreas",
        "FlashMode",
        "WhiteBalance2",
        "WhiteBalanceTemperature",
        "CustomSaturation",
        "ModifiedSaturation",
        "ContrastSetting",
        "SharpnessSetting",
        "NoiseReduction",
        "DistortionCorrection",
        "ShadingCompensation",
        "Gradation",
        "NoiseFilter",
        "DriveMode",
        "PanoramaMode",
        "ImageStabilization",
        "FacesDetected",
        "FaceDetectArea",
    ],
    "Fujifilm MakerNotes": [
        "Version",
        "Quality",
        "Sharpness",
        "WhiteBalance",
        "Saturation",
        "FujiFlashMode",
        "FlashExposureComp",
        "Macro",
        "FocusMode",
        "SlowSync",
        "PictureMode",
        "AutoBracketing",
        "BlurWarning",
        "FocusWarning",
        "ExposureWarning",
    ],
    "Casio MakerNotes": [
        "PreviewImageSize",
        "FirmwareDate",
        "WhiteBalanceBias",
        "ObjectDistance",
        "FlashDistance",
        "RecordMode",
        "ReleaseMode",
        "Quality",
        "FocusMode",
        "HometownCity",
        "BestShotMode",
        "Enhancement",
        "ColorFilter",
        "ArtMode",
        "BracketSequence",
    ],
    "Other Manufacturer Data": [
        "SanyoQuality",
        "Macro",
        "DigitalZoom",
        "DataDump",
        "BWMode",
        "LensDistortionParams",
        "CameraType",
        "PrintIMVersion",
        "FirmwareVersion",
        "RedGain",
        "BlueGain",
        "GreenGain",
        "OffsetTime*",
        "RunTime*",
        "AEStable",
        "AETarget",
        "AEAverage",
        "AFStable",
        "AccelerationVector",
        "TriggerMode",
        "Sequence",
        "EventNumber",
        "MoonPhase",
        "AmbientTemperature*",
        "InfraredIlluminator",
        "MotionSensitivity",
        "BatteryVoltage",
        "UserLabel",
        "Warning*",
        "MakerNote*",
    ],
}


def get_fields(option_name, available_tags=None):
    patterns = METADATA_OPTIONS.get(option_name)

    if patterns is None:
        return None

    if available_tags is None:
        return list(patterns)

    matched = set()
    for pat in patterns:
        if any(ch in pat for ch in "*?[]"):
            for t in available_tags:
                if fnmatch.fnmatchcase(t, pat):
                    matched.add(t)
        else:
            if pat in available_tags:
                matched.add(pat)

    return list(matched)


def list_options():
    return list(METADATA_OPTIONS.keys())



import sys
import importlib.resources

from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication, QMessageBox

from .core import (
    gui, process, exiftool
)


def process_images(filenames, selected_options):
    return process.process_images(filenames, selected_options)

def main():
    app = QApplication(sys.argv)
    app.setApplicationName("MetaClean")
    app.setApplicationDisplayName("MetaClean")
    
    icon_files = importlib.resources.files("metaclean.assets")
    icon_path = icon_files / "icon128.png"
    with importlib.resources.as_file(icon_path) as icon_file:
        app.setWindowIcon(QIcon(str(icon_file)))
    
    if exiftool.check_installed():        
        window = gui.MetaClean(process_images=process_images)
        window.show()
        
        app.exec()
    else:
        print("\033[31mPlease install ExifTool before running this program.\033[0m")
        QMessageBox.critical(
            None,
            "ExifTool Not Found",
            "Please install ExifTool before running this program."
        )
        sys.exit(1)
    
if __name__ == "__main__":
    main()


import sys
from . import exiftool, options

def process_images(filenames, selected_options):
    errors = 0
    for path in filenames:
        try:
            metadata = list(exiftool.get_metadata(path).keys())
        except Exception as e:
            print(f"\033[31mFailed to read metadata for {path}: {e}\033[0m", file=sys.stderr)
            continue
        
        delete_all = False
        to_delete = set()
        for opt in selected_options:
            fields = options.get_fields(opt, metadata)
            if fields is None:
                delete_all = True
                break
            to_delete.update(fields)

        if delete_all:
            exiftool.delete_metadata(path, all=True)
            print(f"Deleted all metadata from: {path}")
        else:
            if not to_delete:
                print(f"No matching tags to delete for {path}")
                continue
            result = exiftool.delete_metadata(path, all=False, properties=list(to_delete))
            if not result:
                errors += 1
            print(f"Deleted {len(to_delete)} tags from: {path}")
            
    return errors


# TODO: Multi-Language support

from PyQt5.QtGui import QIcon, QPixmap
from PyQt5.QtCore import QSize, Qt, pyqtSignal, QTimer
from PyQt5.QtWidgets import (
    QMainWindow,
    QPushButton,
    QLabel,
    QFileDialog,
    QVBoxLayout,
    QHBoxLayout,
    QWidget,
    QListWidget,
    QListWidgetItem,
    QMessageBox,
    QProgressDialog,
    QApplication
)

from . import options

class ClickableLabel(QLabel):
    clicked = pyqtSignal()
    
    def mousePressEvent(self, event):
        self.clicked.emit()
        super().mousePressEvent(event)

class PreviewWindow(QWidget):
    def __init__(self, image_path):
        super().__init__()
        
        self.setWindowFlag(Qt.FramelessWindowHint)
        self.setWindowTitle("Image Preview")
        
        layout = QVBoxLayout()
        image_label = ClickableLabel()
        image_preview = QPixmap(image_path)
        image_preview = image_preview.scaled(
            400, image_preview.height(), Qt.KeepAspectRatio, Qt.SmoothTransformation
        )
        image_label.setPixmap(image_preview)
        image_label.clicked.connect(self.close)
        layout.addWidget(image_label)
        
        self.setLayout(layout)

class MetaClean(QMainWindow):
    def __init__(self, process_images=None):
        super().__init__()
        
        self.filenames = []
        self.preview_windows = []
        self.process_images = process_images
        
        self.setWindowTitle("MetaClean")
        self.setMinimumSize(QSize(550, 400))
        
        self.setup_ui()
        
    def setup_ui(self):
        # create layouts
        layoutV0 = QVBoxLayout()
        layoutH = QHBoxLayout()
        layoutV1 = QVBoxLayout()
        layoutV2 = QVBoxLayout()

        layoutH.setSpacing(10)

        description = QLabel("Select images and choose metadata to remove")
        layoutV0.addWidget(description, alignment=Qt.AlignCenter)

        # left side
        # add widgets
        self.file_list = QListWidget()
        self.file_list.itemDoubleClicked.connect(self.image_preview)
        layoutV1.addWidget(self.file_list)

        # add buttons
        button_layout = QHBoxLayout()
        add_button = QPushButton("Add Images")
        add_button.clicked.connect(self.add_files)
        remove_button = QPushButton("Remove selected")
        remove_button.clicked.connect(self.remove_selected)

        button_layout.addWidget(add_button)
        button_layout.addWidget(remove_button)
        
        # ensure buttons are the same size
        max_width = max(add_button.sizeHint().width(), remove_button.sizeHint().width())
        add_button.setMinimumWidth(max_width)
        remove_button.setMinimumWidth(max_width)

        layoutV1.addLayout(button_layout)

        # right side
        self.metadata_list = QListWidget()
        for name in options.list_options():
            item = QListWidgetItem(name)
            item.setFlags(item.flags() | Qt.ItemIsUserCheckable)
            item.setCheckState(Qt.Unchecked)
            self.metadata_list.addItem(item)
            
        layoutV2.addWidget(self.metadata_list)
        
        # continue button
        continue_button = QPushButton("Continue")
        continue_button.clicked.connect(self.on_continue)
        layoutV2.addWidget(continue_button, alignment=Qt.AlignBottom)

        # add layouts together
        layoutH.addLayout(layoutV1, 1)
        layoutH.addLayout(layoutV2, 1)

        layoutV0.addLayout(layoutH)

        container = QWidget()
        container.setLayout(layoutV0)

        self.setCentralWidget(container)
        
    def add_files(self):
        files, _ = QFileDialog.getOpenFileNames(
            self, "Select Images", "", "Images (*.png *.jpg *.jpeg *.bmp *.gif *.tif *.tiff *.webp)"
        )
        for file in files:
            if file not in self.filenames:
                self.filenames.append(file)
                item = QListWidgetItem()
                pixmap = QPixmap(file)
                if not pixmap.isNull():
                    icon = QIcon(pixmap.scaled(32, 32, Qt.KeepAspectRatio, Qt.SmoothTransformation))
                    item.setIcon(icon)
                item.setText(file)
                self.file_list.addItem(item)

    def remove_selected(self):
        for item in self.file_list.selectedItems():
            self.filenames.remove(item.text())
            self.file_list.takeItem(self.file_list.row(item))

    def image_preview(self, item):
        preview = PreviewWindow(item.text())
        preview.show()
        self.preview_windows.append(preview)

    def get_selected_metadata(self):
        checked = []
        for i in range(self.metadata_list.count()):
            item = self.metadata_list.item(i)
            if item.checkState() == Qt.Checked:
                checked.append(item.text())
        return checked

    def on_continue(self):
        selected_meta = self.get_selected_metadata()
        if selected_meta:
            if self.filenames:
                reply = QMessageBox.question(
                    self,
                    "Confirm Deletion",
                    "Do you really want to continue and delete the selected metadata from your images?",
                    QMessageBox.Yes | QMessageBox.No,
                    QMessageBox.No
                )
                if reply == QMessageBox.Yes:
                    if self.process_images:
                        # TODO: Run in separate thread
                        progress = QProgressDialog("Processing images...", "Cancel", 0, 0, self)
                        progress.setWindowModality(Qt.WindowModal)
                        progress.show()
                        
                        QApplication.processEvents()
                        errors = self.process_images(self.filenames, selected_meta)
                        
                        progress.close()
                        if errors:
                            QMessageBox.warning(
                                self,
                                "Processing Warnings",
                                "Some warnings were encountered during processing. This may occur when certain metadata cannot be removed due to permanent tags."
                            )
                        else:
                            QMessageBox.information(
                                self,
                                "Processing Complete",
                                f"Successfully processed {len(self.filenames)} images."
                            )
            else:
                QMessageBox.warning(
                    self,
                    "No Images Selected",
                    "Select images to continue."
                )
        else:
            QMessageBox.warning(
                self,
                "Nothing to remove",
                "Choose the metadata to remove to continue."
            )